﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.EvolutionTracking;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public class SymbolicDataAnalysisExpressionAfterManipulatorOperator : AfterManipulatorOperator<ISymbolicExpressionTree> {
    private readonly SymbolicExpressionTreeNodeSimilarityComparer comparer;

    public SymbolicDataAnalysisExpressionAfterManipulatorOperator() {
      comparer = new SymbolicExpressionTreeNodeSimilarityComparer {
        MatchVariableNames = true,
        MatchVariableWeights = true,
        MatchConstantValues = true
      };
    }

    public override IOperation Apply() {
      var vChild = (IGenealogyGraphNode<ISymbolicExpressionTree>)GenealogyGraph.GetByContent(ChildParameter.ActualValue);
      var nodesBefore = (List<ISymbolicExpressionTreeNode>)vChild.InArcs.First().Data;
      var nodesAfter = ChildParameter.ActualValue.IterateNodesPrefix().ToList();
      IFragment<ISymbolicExpressionTreeNode> fragment = null;

      for (int i = 0; i < Math.Min(nodesAfter.Count, nodesBefore.Count); ++i) {
        var a = nodesAfter[i];
        var b = nodesBefore[i];

        bool found = false;
        if (comparer.Equals(a, b)) {
          int m = 0;
          for (int j = 0; j < a.SubtreeCount; ++j) {
            if (!AreSimilar(a.GetSubtree(j), b.GetSubtree(j), comparer)) { ++m; }
            if (m > 1) {
              found = true;
              break;
            }
          }
          if (m == 0) {
            // everything is similar so we skip the whole subtree
            i += a.GetLength();
          }
          if (!found) continue;
        }
        fragment = new Fragment<ISymbolicExpressionTreeNode> {
          Root = a,
          Index1 = i,
          Index2 = i
        };
        break;
      }

      vChild.InArcs.First().Data = fragment;
      return base.Apply();
    }

    private bool AreSimilar(ISymbolicExpressionTreeNode a, ISymbolicExpressionTreeNode b, SymbolicExpressionTreeNodeSimilarityComparer comp) {
      int length = a.GetLength();
      return comparer.Equals(a, b) && length.Equals(b.GetLength()) && SymbolicExpressionTreeMatching.Match(a, b, comp) == length;
    }
  }
}
